/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_INCLUDE_FGGL_APP_HPP
#define FGGL_INCLUDE_FGGL_APP_HPP

#define ASSERT_MSG(exp, msg) assert(((void)(msg), exp))

#include <cassert>
#include <string>
#include <memory>
#include <unordered_map>

#include "fggl/display/glfw/window.hpp"
#include <fggl/gfx/paint.hpp>
#include <fggl/util/states.hpp>
#include "fggl/modules/manager.hpp"

namespace fggl {

	class App;

	class AppState;

	using Identifer = std::string;
	using AppMachine = util::StateMachine<AppState, Identifer>;

	class AppState {
		public:
			/**
			 * Create an Application State.
			 *
			 * A state is responsible for managing user interaction with the app. When created, the appstate
			 * is passed a reference to the application that owns it. The lifetime of the state is bounded
			 * by the lifetype of this object.
             *
             * Ie. the lifetime of reference to the App is at least that of the AppState.
			 *
			 * @param owner a non-owned reference to the owner of the state.
			 */
			explicit AppState(App &owner) : m_owner(owner) {}

			virtual ~AppState() = default;

			/**
			 * Update the underlying model of this state.
			 *
			 * States should not assume that one update means one render call, as the game loop may issue
			 * multiple updates per render or vice-versa depending on requriements. Update is intended for
			 * dispatching game-system related infomation.
			 */
			virtual void update(float dt) = 0;

			/**
			 * Perform actions neccerary for rendering the scene.
			 *
			 * When this method is invoked it is safe to assume that rendering of some form will take place
			 * after it returns and the rendering state should be updated to reflect this. The rendering
			 * environment will be passed in as an argument.
			 *
			 * It is not safe to assume the render target will always be the same, as the scene may be
			 * rendered in mutliple passes (eg, for VR requirements).
			 */
			virtual void render(gfx::Graphics &paint) = 0;

            /**
             * Notify the State it has just been switched to.
             *
             * This state should perform any setup it requires (eg, registering callbacks, requesting 
             * resources, etc...).
             *
             * FIXME: This should probably be RAII
             */
			virtual void activate() {}

            /**
             * Notify the State it is about to be switched from.
             *
             * This state should perform any cleanup it requires (eg, unregistering callbacks, freeing
             * resources, etc...).
             *
             * FIXME: This should probably be RAII
             */
			virtual void deactivate() {}

			[[nodiscard]] inline auto owner() const -> App& {
				return m_owner;
			}

		protected:
			App &m_owner;
	};

    /*! \class App app.hpp fggl/app.hpp
     *
     * Main entrypoint to the game framework.
     */
	class App {
		public:
            /**
             * Create an instance of an application, with a set of modules and name.
             *
             * The name is used to derrive verious settings, such as the location of same games and user
             * configuration and content.
             */
			explicit App(modules::Manager *manager, const Identifer &name);

            /**
             * Create an instance of an application, with a set of modules and name.
             *
             * This version of the constructor allows explicitly setting the name used for user-data rather
             * than derriving it from the application name. This is useful if you want to use a shortened
             * version of your application name for configuration.
             */
			App(modules::Manager *manager, const Identifer &name, const Identifer &folderName);

			// class is non copy-able
			App(const App &app) = delete;
			App(const App &&app) = delete;

			auto operator=(const App &other) -> App & = delete;
			auto operator=(App &&other) -> App & = delete;

            /**
             * Set the currently active window.
             *
             * FIXME: this is a nasty hack to get round the setup order for windows, graphics APIs and
             * screen refreshes. Should be fixed with observer pattern.
             */
			inline void setWindow(display::Window *window) {
				m_window = window;
			}

			/**
			 * Perform main game loop functions.
             *
             * You should pass in argc and argv used to invoke main to this method. At the moment it does
             * not use these, but in the future they will be used to provide game-agnostic options.
			 */
			auto run(int argc, const char **argv) -> int;

            /**
             * Register a new state with the application.
             * 
             * States are intended to be a block of code which can be started, will execute a single frame
             * of work, and can be invoked repeatedly. For example, a main menu, options menu, 
             * single-player game and multi-player game might be all be suitable to be states. When a state
             * change is requested, the currently running state is stopped and garabage collected.
             *
             * This is similar to the concept of an 'activity' in Android.
             */
			template<typename T>
			auto addState(const Identifer &name) -> T * {
				static_assert(std::is_base_of<AppState, T>::value, "States must be AppStates");
				return m_states.put<T>(name, *this);
			}

            /**
             * Request the game changes states.
             *
             * This will be executed at the next opporunity (most likley next iteration of the game loop).
             * Identifer should be an identifier used when calling addState, and the state should previouslly
             * been registered using addState.
             */
			inline void change_state(const Identifer &name) {
				m_expectedScene = name;
			}

            /**
             * Return the currently active state (the state that is currently executing).
             */
			inline auto active_state() const -> AppState & {
				return m_states.active();
			}

            /**
             * Get a pointer to a service.
             *
             * This is the primary way in which states can get access to resources they require.
             *
             * returns nullptr if the service does not exist, or cannot be provided.
             */
			template<typename T>
			inline auto service() -> T * {
				try {
					return m_subsystems->template get<T>();
				} catch (std::out_of_range &e) {
					debug::log(debug::Level::error, "Service not found: {}", T::service.get());
					return nullptr;
				}
			}

			inline auto services() -> modules::Services& {
				return m_subsystems->services();
			}

			inline auto running() const -> bool {
				return m_running;
			}

			inline void running(bool state) {
				m_running = state;
			}

		private:
			bool m_running;
			display::Window *m_window;
			AppMachine m_states;
			Identifer m_expectedScene;
			modules::Manager *m_subsystems;
	};

}

#endif
