/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 19/11/22.
//

#ifndef FGGL_ASSETS_PACKED_PACKED_HPP
#define FGGL_ASSETS_PACKED_PACKED_HPP

#include <cstdint>
#include <cstdio>
#include <memory>

#include "fggl/assets/types.hpp"
#include "fggl/assets/packed/direct.hpp"

/**
 * Packed file reader.
 *
 * Read assets stored as sequential [header,data] blocks. This reader does not care about dependencies, it assumes this
 * was handled before storage (ie, asset dependencies are assumed to be stored before the asset that relies on them).
 * The caller is also responsible for ensuring that assets in other archives are already loaded by the time the system
 * assembles the composite assets into something usable.
 *
 * If either of these constraints are violated, the results are undefined.
 */
namespace fggl::assets {

	struct Header {
		uint64_t name;
		uint64_t type;
		std::size_t size;
	};

	bool read_header(std::FILE* stream, Header* header) {
		constexpr auto headerSize = sizeof(Header);
		auto readBytes = std::fread(header, headerSize, 1, stream);
		return readBytes == headerSize;
	}

	bool read_data(std::FILE* stream, void* block, std::size_t size) {
		auto readBytes = std::fread(block, size, 1, stream);
		return readBytes == size;
	}

	void read_archive(RawCheckin* checkin, std::FILE* stream) {
		while ( !std::feof(stream) ) {
			Header header;
			bool headRead = read_header(stream, &header);
			if ( headRead && header.size != 0 ) {
				// header has data
				void* memBlock = std::malloc( header.size );
				bool valid = read_data( stream, memBlock, header.size );

				// read the data, check it in
				if (valid) {
					MemoryBlock block{
						.data = (std::byte*)memBlock,
						.size = header.size
					};
					checkin->check(header.name, header.type, block);
				}
			}
		}
	}
}

#endif //FGGL_ASSETS_PACKED_PACKED_HPP
