/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 18/12/22.
//

#ifndef FGGL_FGGL_GRID_LAYOUT_HPP
#define FGGL_FGGL_GRID_LAYOUT_HPP

#include "fggl/math/types.hpp"
#include "fggl/grid/hexagon.hpp"

namespace fggl::grid {

	// factor out the call to sqrt so the matrices can be constexpr
	constexpr float M_SQRT_3 = 1.73205080757F;
	constexpr float M_SQRT_3_OVER_2 = M_SQRT_3 / 2.0F;

	constexpr math::mat2 MAT_HEX_POINTY{
		M_SQRT_3, 0.0F,
		M_SQRT_3_OVER_2, 3.F / 2.F
	};

	constexpr math::mat2 MAT_HEX_FLAT{
		3.F / 2.F, M_SQRT_3_OVER_2,
		0.0F, M_SQRT_3
	};

	struct Orientation {
		math::mat2 m_forward;
		math::mat2 m_backward;
		int m_angle;

		inline Orientation(int angle, math::mat2 forward) : m_forward(forward), m_backward(glm::inverse(forward)), m_angle(angle) {};

		static inline Orientation make_pointy() {
			return { 30, MAT_HEX_POINTY};
		}

		static inline Orientation make_flat() {
			return { 0, MAT_HEX_FLAT };
		}
	};


	constexpr int HEX_SIDES = 6;
	constexpr int DEG_PER_HEX_SIDE = 360 / HEX_SIDES; // 60 degrees per side

	struct Layout {
		Orientation m_orientation;
		math::vec2 m_size;
		math::vec2 m_origin;

		Layout(Orientation orientation, math::vec2 size, math::vec2 origin) : m_orientation(orientation), m_size(size), m_origin(origin){}
		Layout(Orientation orientation, float size) : m_orientation(orientation), m_size(size, size), m_origin() {}

		inline void translate(float dx, float dy){
			m_origin.x += dx;
			m_origin.y += dy;
		}

		[[nodiscard]]
		inline math::vec2 origin() const {
			return m_origin;
		}

		[[nodiscard]]
		inline math::vec2 size() const {
			return m_size;
		}

		[[nodiscard]]
		inline math::vec2 toScreen(IntHex gridPos) const {
			const math::vec2 hexPoint{gridPos.q(), gridPos.r()};
			auto point = (m_orientation.m_forward * hexPoint) * m_size;
			return point + m_origin;
		}

		[[nodiscard]]
		inline FloatHex toGrid(math::vec2 screen) const {
			auto point = (screen - m_origin) / m_size;
			auto hexPos = m_orientation.m_backward * point;
			return {hexPos.x, hexPos.y};
		}

		[[nodiscard]]
		inline math::vec2 cornerOffset(int corner) const {
			const int angInc = DEG_PER_HEX_SIDE * corner;
			const float angle = (angInc + m_orientation.m_angle) * ( fggl::math::PI / 180.0F);
			return {
				m_size.x * cosf(angle),
				m_size.y * sinf(angle)
			};
		}

		void paintHex(fggl::gfx::Paint& paint, IntHex pos, math::vec3 colour, math::vec2 offset) const {
			const auto hexScreenCenter = toScreen(pos);

			gfx::Path2D path({0,0});
			path.colour(colour);

			for (int i=0; i < HEX_SIDES; ++i) {
				auto cornerPos =  hexScreenCenter + cornerOffset(i);
				if ( i == 0) {
					path.moveTo(cornerPos + offset);
				} else {
					path.pathTo(cornerPos + offset);
				}
			}
			paint.stroke(path);
		}


	};

}

#endif //FGGL_FGGL_GRID_LAYOUT_HPP
