/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/06/22.
//

#ifndef FGGL_MODULES_MANAGER_HPP
#define FGGL_MODULES_MANAGER_HPP

#include "fggl/modules/module.hpp"
#include "fggl/debug/logging.hpp"
#include "fggl/ds/graph.hpp"

#include <cassert>
#include <queue>
#include <vector>
#include <map>
#include <stack>
#include <set>
#include <iostream>

namespace fggl::modules {

	/**
	 * Store and initialise modules present in the engine.
	 *
	 * This class is responsible for keeping track of which modules the library user has requested, and ensuring that
	 * their dependencies are loaded in the correct order. Once the dependency graph has been built and instances
	 * created, it is responsible for providing references to these initialised classes.
	 */
	class Manager {
		public:
			Manager() = default;

			template<ServiceType T>
			class Service {
				public:
					inline Service(Manager* manager) : m_manager(manager) {}

					inline T* operator->() {
						if ( m_ptr == nullptr ) {
							m_ptr = m_manager->get<T>();
						}
						return m_ptr;
					}

				private:
					Manager* m_manager;
					std::shared_ptr<T> m_ptr;
			};

			inline void addVirtual(const Config &config) {
				assert(!m_locked);

				m_modules[config.name] = config;
				for (const auto &service : config.provides) {
					m_serviceProviders[service] = config.name;
				}
			}

			template<ModuleType T>
			void use() {
				assert(!m_locked);

				Config config{.name = T::name, .provides = {}, .depends = {}};
				for (auto service : T::provides) {
					config.provides.push_back(service);
				}

				for (auto service : T::depends) {
					config.depends.push_back(service);
				}
				config.factory = T::factory;

				addVirtual(config);
			}

            // FIXME this should be in cpp file
			bool buildGraph() {
				// resolve links between modules
				for (auto &moduleItr : m_modules) {
					if (moduleItr.second.depends.empty()) {
						m_dependencies.addVertex(moduleItr.first);
						continue;
					}

					for (auto &service : moduleItr.second.depends) {

						auto provider = m_serviceProviders.find(service);
						if (provider == m_serviceProviders.end()) {
							debug::log(debug::Level::warning,
									   "{} depends on {}, but nothing provides it",
									   moduleItr.first,
									   service.get());
							// nothing can provide the service requested, setup is invalid.
							return false;
						}

						m_dependencies.addEdge(moduleItr.first, provider->second);
					}
				}
				return true;
			}

			template<ServiceType T>
			T *get() const {
				assert(m_locked);
				return m_services.template get<T>();
			}

			template<ServiceType T>
			Service<T> getLazy() const {
				assert(m_locked);
				return { this };
			}

            //FIXME this should be in cpp file!
			void resolve() {
				assert( !m_locked );
				if (!buildGraph()) {
					return;
				}

				std::queue<ModuleIdentifier> stack;
				m_dependencies.getOrderRev(stack);

				while (!stack.empty()) {
					auto nextToInit = stack.front();
					debug::log(debug::Level::info, "Initializing {}", nextToInit);

					auto &module = m_modules.at(nextToInit);
					if (module.factory != nullptr) {
						for (auto &service : module.provides) {
							bool result = module.factory(service, m_services);
							if (!result) {
								debug::log(debug::Level::warning,
										   "{} could not create service {}",
										   nextToInit,
										   service.get());
							}
						}
					} else {
						debug::log(debug::Level::warning, "{} has no factory - skipping", nextToInit);
					}
					stack.pop();
				}

				debug::log(debug::Level::info, "engine boot complete");
				m_locked = true;
			}

			inline Services& services() {
				return m_services;
			}

		private:
			bool m_locked = false;
			Services m_services;
			std::map<ModuleIdentifier, Config> m_modules;
			ds::DirectedGraph<ModuleIdentifier> m_dependencies;
			std::map<ServiceName, ModuleIdentifier> m_serviceProviders;

	};

} // namespace fggl::modules

#endif //FGGL_MODULES_MANAGER_HPP
