/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 25/03/23.
//

#ifndef FGGL_MODULES_SERVICE_HPP
#define FGGL_MODULES_SERVICE_HPP

#include <map>
#include <memory>

#include "fggl/util/safety.hpp"

namespace fggl::modules {

	using ServiceName = util::OpaqueName<std::string_view, struct ModuleServiceTag>;
	constexpr ServiceName make_service(const std::string_view name) {
		return ServiceName::make(name);
	}

	template<typename T>
	concept ServiceType = requires(T* type) {
		{ T::service };
	};

	template<typename T, typename U>
	concept Derived = std::is_base_of<U, T>::value;

	class Services {
		public:
			template<ServiceType Svc, Derived<Svc> Impl, typename ...Args>
			void bind(Args... args) {
				m_services[Svc::service] = std::make_shared<Impl>(args...);
			}

			template<ServiceType Svc, typename ...Args>
			Svc *create(Args... args) {
				auto svc = std::make_shared<Svc>(args...);
				m_services[Svc::service] = svc;
				return svc.get();
			}

			template<ServiceType Svc>
			void provide(std::shared_ptr<Svc> service) {
				m_services[Svc::service] = service;
			}

			template<ServiceType S>
			S *get() const {
				auto serviceWrapper = m_services.at(S::service);
				auto ptr = std::static_pointer_cast<S>(serviceWrapper);
				return ptr.get();
			}

		private:
			std::map<ServiceName, std::shared_ptr<void>> m_services;
	};

} // namespace fggl::modules



#endif //FGGL_MODULES_SERVICE_HPP
